(* Expressions *)

type exp = Int of int
         | Ide of string
         | App of string * exp
         | Add of exp * exp
         | Sub of exp * exp
         | Mul of exp * exp
         | Ifz of exp * exp * exp
         | Eq of exp * exp
         | Leq of exp * exp
         | Not of exp
         | And of exp * exp
         | Or of exp * exp

type defin = Funz of string * string * exp;;

type prog = Program of (defin list) * exp;;

(* run time *)

exception Ahhhh;;

(* Global env -- default *)
let fenv0 = fun x -> raise Ahhhh;;

(* Local env -- default *)
let env0 = fenv0;;

(* Update Env *)
let ext env (x:string) v = fun y -> if x=y then v else env y

(*
env, fenv : ’a -> ’b
ext : (’a -> ’b) -> ’a -> ’b -> ’a -> ’b
env : string -> int
fenv : string -> (string * exp) 
*)

(* Eval
   NB:  bollean  0 (false) -- 1 (true) *)

let rec eval (e:exp) env fenv =
match e with
  Int i -> i
| Ide s -> env s
| App (s,e2) -> let (par,body) = (fenv s)
      in (eval body (ext env par (eval e2 env fenv)) fenv)
| Add (e1,e2) -> (eval e1 env fenv)+(eval e2 env fenv)
| Sub (e1,e2) -> (eval e1 env fenv)-(eval e2 env fenv)
| Mul (e1,e2) -> (eval e1 env fenv)*(eval e2 env fenv)
| Not e1 -> if ((eval e1 env fenv) = 0)  then 1 else 0 
| Or (e1,e2) -> if (eval e1 env fenv) = 0 then (eval e2 env fenv) else 1 
| And (e1,e2) -> if (eval e1 env fenv) != 0 then (eval e2 env fenv) else 0
| Eq (e1,e2) -> if ((eval e1 env fenv) = (eval e2 env fenv)) then 1 else 0 
| Leq (e1,e2) -> if ((eval e1 env fenv) <= (eval e2 env fenv)) then 1 else 0
| Ifz (e1,e2,e3) -> if (eval e1 env fenv) = 1
                     then (eval e2 env fenv)
                     else (eval e3 env fenv);;

(* valutazione di dichiarazione: restituisce un ambiente globale *)
let rec dval (decls:defin list) = match decls with
    [] -> fenv0
    |  Funz (fname,par,body)::rest -> ext (dval rest) fname (par, body);;

(* eval program *)
let peval (p:prog) =
    match p with
    Program (decls,expr) ->
        let fenv = dval(decls) in eval expr env0 fenv;;

(* =============================  TESTS  ================= *)


let p1 = Program ([], Add(Int 4, Int 5));;

peval p1;;


let p2 = Program ([Funz("succ","x",Add(Ide "x",Int 1))], Add(App("succ", Int 4), Int 5));;

peval p2;;


let p3 = Program ([Funz("tria","x",Ifz(Eq(Ide "x",Int 0),Int 5,Add(Ide "x",App("tria",Sub(Ide "x",Int 1)))))], App("tria", Int 4));;

peval p3;;


let p4 = Program ([Funz("fact","x",Ifz(Leq(Ide "x",Int 1),Int 1,Mul(Ide "x",App("fact",Sub(Ide "x",Int 1)))))], App("fact", Int 3));;

peval p4;;


let ptest = 
Program 
([Funz ("sub1", "n", Sub(Ide "n", Int 1));
  Funz ("fib", "n", 
       Ifz(Or(Eq(Ide "n",Int 0), Eq(Ide "n", Int 1)), Ide "n",
	     Add(App("fib",App("sub1", Ide "n")),
		    App("fib",App("sub2", Ide "n")))));
  Funz ("sub2", "m", Sub(Ide "m", Int 2))],
  App("fib", Int 7));;

peval ptest;;





